function f = optimal_tax(x,T,a0,A0,z,Z,G,TAUK);

GLOBALS 

tauK0   = TAUK(1);
TAUK    = TAUK(2:end);
N       = length(TAUK);
f       = zeros(8*T+2+N,size(x,2));
FILL1   = ones(T,1);
FILL2   = ones(1,size(x,2));

x(1:5*T,:) = exp(x(1:5*T,:));
c          = x(1:T,:);
C          = x(T+1:2*T,:);
h          = x(2*T+1:3*T,:);
H          = x(3*T+1:4*T,:);
nextK      = x(4*T+1:5*T,:);
nu         = x(5*T+1:6*T,:);
rho        = x(6*T+1:7*T,:);
xsi        = x(7*T+1:8*T,:);
lambda     = x(8*T+1,:);
Lambda     = x(8*T+2,:);
psi        = x(8*T+3:8*T+2+N,:);               % multiplier on constr that tauK = TAUK

K0      = A0 - Dss;
K       = [K0*FILL2; nextK(1:end-1,:)];
L       = Z*H;

v1   = U1(c,h);
v2   = U2(c,h);
v11  = U11(c,h);
v12  = U12(c,h);
v22  = U22(c,h);
V1   = U1(C,H);
V2   = U2(C,H);
V11  = U11(C,H);
V12  = U12(C,H);
V22  = U22(C,H);

Wc   = v1 + (FILL1*lambda).*(v1 + v11.*c + v12.*h);
WC   = (FILL1*Lambda).*(V1 + V11.*C + V12.*H);
Wh   = v2 + (FILL1*lambda).*(v12.*c + v2 + v22.*h);
WH   = (FILL1*Lambda).*(V12.*C + V2 + V22.*H);

Y    = K.^theta .* L.^(1-theta);
FK   = theta * Y./K;
FL   = (1-theta) * Y./L;
FKK  = theta*(theta-1) * Y./K./K;
FKL  = theta*(1-theta) * Y./K./L;
R0   = 1 + (1-tauK0) .* (FK(1,:) - delta);
RH0  = (1-tauK0) * Z .* FKL(1,:);

R       = V1(1:end-1,:) ./ (bta*V1(2:end,:));
tauK    = 1+(1-R)./(FK(2:end,:)-delta);
psi     = [psi; zeros(T-N-1,size(x,2))];
TAUK    = [TAUK; zeros(T-N-1,1)]*FILL2;
GC      = V1(1:end-1,:) .* V11(2:end,:) ./ (bta*V1(2:end,:).^2);                            % G = tauK - TAUK; 
GH      = (1-TAUK)*Z.*FKL(2:end,:) + V1(1:end-1,:) .* V12(2:end,:) ./ (bta*V1(2:end,:).^2);
GK      = (1-TAUK).*FKK(2:end,:);
GClag   = -V11(1:end-1,:) ./ (bta*V1(2:end,:)); GClag = [GClag; GClag(end,:)];
GHlag   = -V12(1:end-1,:) ./ (bta*V1(2:end,:)); GHlag = [GHlag; GHlag(end,:)];


% =====================================================================
% Foc{c}
% =====================================================================
f(1:T,:) = Wc + ...
           rho.*v11.*[V1(2:end,:); V1(end,:)] + ...
           xsi.*(z*v11.*V2 - Z*V1.*v12);      
f(1,:)   = f(1,:) - lambda .* v11(1,:) .* R0 * a0 / (1+tauC);
f(2:T,:) = f(2:T,:) -  rho(1:end-1,:).*V1(1:end-1,:).*v11(2:end,:)/bta;
      
% =====================================================================
% Foc{C}
% =====================================================================
f(T+1:2*T,:) = WC - nu - ...
               rho.*V11.*[v1(2:end,:); v1(end,:)] + ...
               xsi.*(z*v1.*V12 - Z*V11.*v2);
f(T+1,:)     = f(T+1,:) - Lambda .* V11(1,:) .* R0 * A0 / (1+tauC) - ...
               bta*GClag(1,:).*psi(1,:);
f(T+2:2*T,:) = f(T+2:2*T,:) + rho(1:end-1,:).*v1(1:end-1,:).*V11(2:end,:)/bta - ...
               GC.*psi - ...
               bta*GClag(2:end,:).*[psi(2:end,:); psi(end,:)];


% =====================================================================
% Foc{h}
% =====================================================================
f(2*T+1:3*T,:) = Wh + rho.*v12.*[V1(2:end,:); V1(end,:)] + ...
                 xsi.*(z*v12.*V2 - Z*V1.*v22);
f(2*T+1,:)     = f(2*T+1,:) - lambda .* v12(1,:).*R0 * a0 / (1+tauC);
f(2*T+2:3*T,:) = f(2*T+2:3*T,:) - rho(1:end-1,:).*V1(1:end-1,:).*v12(2:end,:)/bta;

% =====================================================================
% Foc{H}
% =====================================================================
f(3*T+1:4*T,:) = WH + nu*Z.*FL - rho.*V12.*[v1(2:end,:); v1(end,:)] + ...
                 xsi.*(z*v1.*V22 - Z*V12.*v2);
f(3*T+1,:)     = f(3*T+1,:) - lambda .* (v1(1,:).*RH0) * a0  / (1+tauC) - ...
                              Lambda .* (V12(1,:).*R0 + V1(1,:).*RH0) * A0 / (1+tauC) - ...
                 bta*GHlag(1,:).*psi(1,:);
f(3*T+2:4*T,:) = f(3*T+2:4*T,:) + rho(1:end-1,:).*v1(1:end-1,:).*V12(2:end,:)/bta - ...
                 GH.*psi - ...
                 bta*GHlag(2:end,:).*[psi(2:end,:); psi(end,:)];


% =====================================================================
% Foc nextK
% =====================================================================
f(4*T+1:5*T-1,:) = nu(2:end,:).*(FK(2:end,:)+1-delta) - nu(1:end-1,:)/bta - ...
                   GK.*psi;
f(5*T,:)         = nextK(end,:) - nextK(end-1,:);

% =====================================================================
% Resource constr
% =====================================================================

f(5*T+1:6*T,:) = C + G + nextK - Y - (1-delta)*K;

% =====================================================================
% Same capital income tax rate on all individuals
% =====================================================================

f(6*T+1:7*T,:) = v1.*[V1(2:end,:); V1(end,:)] - ...
                 V1.*[v1(2:end,:); v1(end,:)];

% =====================================================================
% Same labor income tax rate on all individuals
% =====================================================================

f(7*T+1:8*T,:) = z*v1.*V2 - Z*V1.*v2;

% =====================================================================
% IC
% =====================================================================
ic         = (bta.^(0:T-1)'*FILL2) .* (v1.*c + v2.*h);
f(8*T+1,:) = sum(ic) + bta*ic(end,:)/(1-bta) - v1(1,:).*R0*a0/(1+tauC);

Ic         = (bta.^(0:T-1)'*FILL2) .* (V1.*C + V2.*H);
f(8*T+2,:) = sum(Ic) + bta*Ic(end,:)/(1-bta) - V1(1,:).*R0*A0/(1+tauC);

% =====================================================================
% Restriction on tauK
% =====================================================================
f(8*T+3:8*T+2+N,:) = tauK(1:N,:) - TAUK(1:N,:);
